# -*- coding: utf-8 -*-
"""
Created on Thu Jul  6 11:38:34 2023
"""
import diskcache
import os
import pandas as pd
import io
from requests_cache import CachedSession
from requests import Session
from datetime import timedelta
import logging
import numpy as np
import time
from tqdm import tqdm
from pebble import ThreadPool
from pynsee.localdata import get_area_list
from rapidfuzz import fuzz, process
from unidecode import unidecode

from french_cities import DIR_CACHE
from french_cities.utils import init_pynsee, patch_the_patch


logger = logging.getLogger(__name__)


def _process_departements_from_postal(
    df: pd.DataFrame,
    source: str,
    alias: str,
    session: Session = None,
    authorize_duplicates: bool = False,
) -> pd.DataFrame:
    """
    Retrieve departement's code from postoffice code. Adds the result as a new
    column to dataframe under the label 'alias'. Uses the BAN (Base Adresse
    Nationale under the hood) and OpenDataSoft's freemium API in backoffice (
    mostly in case of "Cedex" codes)

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing postal codes
    source : str
        Field containing the postal codes
    alias : str
        Column to store the departements' codes unto
    session : Session, optional
        Web session. The default is None (and will use a CachedSession with
        30 days expiration)
    authorize_duplicates : bool, optional
        If True, authorize duplication of results when multiple results are
        acceptable for a given postcode (for instance, 13780 can result to
        either 13 or 83). If False, duplicates will be removed, hence no
        result will be available. False by default.

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """

    cache_departments = diskcache.Cache(os.path.join(DIR_CACHE, "deps"))

    if not session:
        session = CachedSession(
            cache_name=os.path.join(DIR_CACHE, "find-department"),
            allowable_methods=("GET", "POST"),
            expire_after=timedelta(days=30),
        )

    df["#CachedResult#"] = df[source].apply(cache_departments.get)

    # Download official postcodes dataset from API
    # https://datanova.laposte.fr/datasets/laposte-hexasmal
    url = (
        "https://datanova.laposte.fr/data-fair/api/v1/datasets/"
        "laposte-hexasmal/raw"
    )

    r = session.get(url)
    base = pd.read_csv(
        io.BytesIO(r.content), sep=";", encoding="cp1252", dtype=str
    )
    base = base[["#Code_commune_INSEE", "Code_postal"]]
    base = _process_departements_from_insee_code(
        base, "#Code_commune_INSEE", alias
    )
    base = base[["Code_postal", alias]].rename({"Code_postal": source}, axis=1)
    base = base.drop_duplicates(keep="first")

    ix = df[df["#CachedResult#"].isnull()].index

    result_hexasmal = df.loc[ix].merge(base, on=source, how="inner")
    result_hexasmal = (
        result_hexasmal[[alias, source]].dropna().drop_duplicates()
    )

    ix = df[df["#CachedResult#"].isnull()].index
    postal_codes_ban = (
        df.loc[ix, [source]]
        .merge(
            result_hexasmal[[source]], on=source, how="left", indicator=True
        )
        .query('_merge=="left_only"')
        .drop("_merge", axis=1)
        .drop_duplicates(keep="first")
    )
    if not postal_codes_ban.empty:

        files = [
            (
                "data",
                (
                    "data.csv",
                    postal_codes_ban.to_csv(index=False, encoding="utf8"),
                    "text/csv; charset=utf-8",
                ),
            ),
            ("postcode", (None, source)),
            ("result_columns", (None, source)),
            ("result_columns", (None, "result_context")),
        ]

        with patch_the_patch():
            r = session.post(
                # recherche grâce à l'API de la BAN
                "https://api-adresse.data.gouv.fr/search/csv/",
                files=files,
            )

        if not r.ok:
            raise Exception(
                f"Failed to query BAN's API with {files=} - response was {r}"
            )
        result = pd.read_csv(io.BytesIO(r.content), dtype=str)
        result[alias] = (
            result["result_context"]
            .str.split(",", expand=True)[0]
            .str.strip(" ")
        )
        result = result.drop("result_context", axis=1)
        result_ban = result[result[alias].notnull()].drop_duplicates()
    else:
        result_ban = pd.DataFrame()

    result = pd.concat([result_hexasmal, result_ban]).drop_duplicates()

    ix = df[df["#CachedResult#"].isnull()].index
    postal_codes_cedex = (
        df.loc[ix, [source]]
        .merge(result[[source]], on=source, how="left", indicator=True)
        .query('_merge=="left_only"')
        .drop("_merge", axis=1)
        .drop_duplicates(keep="first")
    )

    # where code is unknown, use Christian Quest Dataset with Cedex codes and
    # OpenDataSoft API (v2.1 contrairement à la doc disponible) en Freemium
    # https://www.data.gouv.fr/fr/datasets/liste-des-cedex/#_
    # https://public.opendatasoft.com/explore/dataset/correspondance-code-cedex-code-insee/information/?flg=fr&q=code%3D68013&lang=fr
    def get(x):
        while True:
            r = session.get(
                # recherche "en masse" grâce à l'API de la BAN
                "https://public.opendatasoft.com/api/explore/v2.1/catalog/datasets/correspondance-code-cedex-code-insee/records",
                params={
                    "select": "insee,libelle,nom_com",
                    "where": f"code={x}",
                    "limit": "10",
                    "offset": "0",
                    "timezone": "UTC",
                    "include_links": "false",
                    "include_app_metas": "false",
                },
            )
            if r.ok:
                break
            elif r.status_code == 429:
                time.sleep(1)
            else:
                logger.warning(
                    f"Error occured on code {x} on OpenDataSoft's API"
                )
                return None
        results = r.json()["results"]

        for dict_ in results:
            dict_.update({source: x})
        return results

    if not postal_codes_cedex.empty:
        logger.info("postal codes unrecognized - maybe Cedex codes")
        args = postal_codes_cedex[source].dropna().tolist()
        result_cedex = []
        with tqdm(
            total=len(args), desc="Querying OpenDataSoft API", leave=False
        ) as pbar:
            with ThreadPool(10) as pool:
                future = pool.map(get, args)
                results_iterator = future.result()
                while True:
                    try:
                        this_result = next(results_iterator)
                        if this_result:
                            result_cedex += this_result
                    except StopIteration:
                        break
                    finally:
                        pbar.update(1)
        result_cedex = pd.DataFrame(result_cedex).drop_duplicates()

        if not result_cedex.empty:
            # Keep only results with valid department
            result_cedex = _process_departements_from_insee_code(
                result_cedex,
                source="insee",
                alias=alias,
                session=session,
            )
            ix = result_cedex[result_cedex[alias].notnull()].index
            result_cedex = result_cedex.loc[
                ix, [source, alias]
            ].drop_duplicates(keep="first")
    else:
        result_cedex = pd.DataFrame()

    result = pd.concat(
        [result_hexasmal, result_ban, result_cedex]
    ).drop_duplicates()

    ix = df[df["#CachedResult#"].isnull()].index
    postal_codes_missing = (
        df.loc[ix, [source]]
        .merge(result[[source]], on=source, how="left", indicator=True)
        .query('_merge=="left_only"')
        .drop("_merge", axis=1)
        .drop_duplicates(keep="first")
    )

    if not postal_codes_missing.empty:
        # Still no results -> assume we can use the first characters of
        # postcode anyway
        last_resort_results = _process_departements_from_insee_code(
            postal_codes_missing,
            source=source,
            alias="dep",
            session=session,
        )
        last_resort_results = last_resort_results.dropna()
    else:
        last_resort_results = pd.DataFrame()

    result = pd.concat(
        [result_hexasmal, result_ban, result_cedex, last_resort_results]
    )

    result = result.drop_duplicates()
    if not authorize_duplicates:
        result = result.drop_duplicates(source, keep=False)

    logger.info("résultat obtenu")

    df = df.merge(result, on=source, how="left").drop_duplicates()
    ix = df[df["#CachedResult#"].notnull()].index
    df.loc[ix, alias] = df.loc[ix, "#CachedResult#"]
    ix = df[df["#CachedResult#"].isnull()].index

    new_cache_values = df.loc[ix, [source, alias]].drop_duplicates()

    # Cache only non-duplicated results!
    new_cache_values = new_cache_values.drop_duplicates(source, keep=False)

    new_cache_values = dict(new_cache_values.values)
    for key, val in new_cache_values.items():
        cache_departments[key] = val

    df = df.drop("#CachedResult#", axis=1)
    return df


def _process_departements_from_insee_code(
    df: pd.DataFrame,
    source: str,
    alias: str,
    session: Session = None,
    authorize_duplicates: bool = False,
) -> pd.DataFrame:
    """
    Compute departement's codes from official french cities codes (COG INSEE).
    Adds the result as a new column to dataframe under the label 'alias'.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing official cities codes
    source : str
        Field containing the official codes (INSEE COG)
    alias : str
        Column to store the departements' codes unto
    session : Session, optional
        **ignored argument, set only for coherence with
        _process_departements_from_postal**
    authorize_duplicates : bool, optional
        **ignored argument, set only for coherence with
        _process_departements_from_postal**
    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """
    init_pynsee()
    deps = get_area_list("departements", date="*").rename(
        {"CODE": "#DEP_CODE#"}, axis=1
    )
    deps = deps[["#DEP_CODE#"]].drop_duplicates(keep="first")

    df[alias] = df[source].str[:2]

    ix = df[df[alias] == "97"].index
    df.loc[ix, alias] = df.loc[ix, source].str[:3]

    # Remove unvalid results (ultramarine collectivity, monaco, ...)
    df = df.merge(deps, left_on=alias, right_on="#DEP_CODE#", how="left")
    df = df.drop(alias, axis=1).rename({"#DEP_CODE#": alias}, axis=1)
    return df


def find_departements(
    df: pd.DataFrame,
    source: str,
    alias: str,
    type_code: str,
    session: Session = None,
    authorize_duplicates: bool = False,
) -> pd.DataFrame:
    """
    Compute departement's codes from postal or official codes (ie. INSEE COG)'
    Adds the result as a new column to dataframe under the label 'alias'.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing official cities codes
    source : str
        Field containing the post or official codes
    alias : str
        Column to store the departements' codes unto
    type_code : str
        Type of codes passed under `alias` label. Should be either 'insee' for
        official codes or 'postcode' for postal codes.
    session : Session, optional
        Web session. The default is None (and will use a CachedSession with
        30 days expiration)
    authorize_duplicates : bool, optional
        If True, authorize duplication of results when multiple results are
        acceptable for a given postcode (for instance, 13780 can result to
        either 13 or 83 ). If False, duplicates will be removed, hence no
        result will be available. False by default.

    Raises
    ------
    ValueError
        If type_code not among "postcode", "insee".

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """
    if type_code not in {"postcode", "insee"}:
        msg = (
            "type_code must be among ('postcode', 'insee') - "
            f"found {type_code} instead"
        )
        raise ValueError(msg)

    init_pynsee()

    df = df.copy()
    if type_code == "postcode":
        func = _process_departements_from_postal
    elif type_code == "insee":
        func = _process_departements_from_insee_code
    return func(df, source, alias, session, authorize_duplicates)


def find_departements_from_names(
    df: pd.DataFrame, label: str, alias: str = "DEP_CODE"
) -> pd.DataFrame:
    """
    Retrieve departement's codes from their names.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing official departement's names
    label : str
        Field containing the label of the departements
    alias : str, optional
        Column to store the departements' codes unto.
        Default is "DEP_CODE"

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """

    init_pynsee()
    candidates = get_area_list("departements")
    candidates = candidates[["CODE", "TITLE"]]
    candidates["TITLE"] = (
        candidates["TITLE"]
        .apply(unidecode)
        .str.upper()
        .str.replace(r"\W+", " ", regex=True)
    )
    candidates = dict(candidates[["TITLE", "CODE"]].values)
    candidates_keys = list(candidates.keys())

    df = df.copy()
    df["FORMATTED"] = (
        df[label]
        .apply(unidecode)
        .str.upper()
        .str.replace(r"\W+", " ", regex=True)
    )

    df[alias] = df["FORMATTED"].apply(
        lambda x: candidates[
            process.extractOne(
                x,
                candidates_keys,
                scorer=fuzz.ratio,
                score_cutoff=80,
            )[0]
        ]
    )
    df = df.drop("FORMATTED", axis=1)

    return df
